<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 5/9/17
 * Time: 8:56 AM
 */

namespace qfq;

require_once(__DIR__ . '/../exceptions/CodeException.php');
require_once(__DIR__ . '/../helper/Logger.php');
require_once(__DIR__ . '/Database.php');

/*
 * Read the extension version number.
 * Read the QFQ database version number: stored in the comment field of table 'Form'. Format:  Version=x.y.z
 * If versions different:
 *   * Read the update array 'DatabaseUpdateData.php'.
 *   * Play all changes from the update array after 'old' upto 'new'.
 *   * Save new QFQ database version in the comment field of table 'Form'
 *
 * In a new QFQ installation, the comment field of table 'Form' is empty. On the first call of QFQ, the version string
 * will be set. Also the 'formEditor.sql' will be played initially.
 *
 */

/**
 * Class DatabaseUpdate
 * @package qfq
 */
class DatabaseUpdate {

    /**
     * @var Database instantiated class
     */
    protected $db = null;

    /**
     * @param Database $db
     */
    public function __construct(Database $db) {
        $this->db = $db;
    }

    /**
     * @return mixed
     * @throws CodeException
     */
    private function getExtensionVersion() {
        $path = __DIR__ . '/../../../ext_emconf.php';
        $_EXTKEY = EXT_KEY;
        $EM_CONF = null;
        if (@file_exists($path)) {
            include $path;

            if (isset($EM_CONF[$_EXTKEY]['version'])) {
                return $EM_CONF[$_EXTKEY]['version'];
            }
        }

        throw new CodeException('Failed to read extension version', ERROR_QFQ_VERSION);
    }

    /**
     * Try to read the QFQ version number from 'comment' in table 'Form'.
     * In a very special situation , there might be table 'form' and 'Form' in the same Database. This should be handled
     *  in the way that the latest version number is the active one.
     *
     * @return bool|string  false if there is no table 'Form' or if there is no comment set in table 'Form'.
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    private function getDatabaseVersion() {

        $arr = $this->db->sql("SHOW TABLE STATUS WHERE Name='Form'", ROW_REGULAR);

        $found = '';
        //
        foreach ($arr as $row) {
            if (isset($row['Comment'])) {
                parse_str($row['Comment'], $arr);
                if (($arr['Version'] ?? '') !== '' AND (version_compare($arr['Version'], $found) == 1)) {
                    $found = $arr;
                }
            } else {
                continue;
            }
        }

        return ($found === '') ? false : $found;
    }

    /**
     * @param $version
     *
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    private function setDatabaseVersion($version) {

        if (is_array($version)) {
            $versionInfo = $version;
        } else {
            $versionInfo = $this->getDatabaseVersion();
            $versionInfo['Version'] = $version;
        }

        $this->db->sql("ALTER TABLE `Form` COMMENT = '" . http_build_query($versionInfo) . "'");

    }

    /**
     *
     * @param string $dbUpdate SYSTEM_DB_UPDATE_ON | SYSTEM_DB_UPDATE_OFF | SYSTEM_DB_UPDATE_AUTO
     *
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    public function checkNupdate($dbUpdate) {

        if ($dbUpdate === SYSTEM_DB_UPDATE_NEVER) {
            return;
        }
        $old = false;
        $new = $this->getExtensionVersion();
        $versionInfo = $this->getDatabaseVersion();
        $old = $versionInfo['Version'];

        if ($dbUpdate === SYSTEM_DB_UPDATE_ALWAYS || ($dbUpdate === SYSTEM_DB_UPDATE_AUTO && $new != $old)) {

            $newFunctionsHash = $this->updateSqlFunctions($versionInfo['functionsHash']);
            if (null !== $newFunctionsHash) {
                $versionInfo['functionsHash'] = $newFunctionsHash;
                $versionInfo['functionsVersion'] = $new;
            }

            $this->dbUpdateStatements($old, $new);
            $this->db->playSqlFile(__DIR__ . '/../../sql/formEditor.sql');

            $qfqLog = $this->db->getQfqLogFile();
            Logger::logMessage(date('Y.m.d H:i:s ') . ": Updated from QFQ version '$old' to '$new'", $qfqLog);

            // Finally write the latest version number.
            $versionInfo['Version'] = $new;
            $this->setDatabaseVersion($versionInfo);
        }

        if ($old === false) {
            // A complete new installation get's some extra tables
            $this->db->playSqlFile(__DIR__ . '/../../sql/customTable.sql');
        }
    }

    /**
     * @param $oldFunctionsHash
     *
     * @return string
     *
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    private function updateSqlFunctions($oldFunctionsHash) {
        if ('no' === $oldFunctionsHash) {
            return null;
        }

        $functionsSql = file_get_contents(__DIR__ . '/../../sql/functions.sql');
        $functionsHash = hash('md5', $functionsSql);

        if ($functionsHash === $oldFunctionsHash) {
            return null;
        }

        $query = str_replace('%%FUNCTIONSHASH%%', $functionsHash, $functionsSql);
        if (stripos($query, 'delimiter')) {
            $errorMsg[ERROR_MESSAGE_TO_USER] = 'Error while updating qfq.';
            $errorMsg[ERROR_MESSAGE_TO_DEVELOPER] = "Error in file functions.sql: The keyword DELIMITER is present " .
                "in functions.sql, this usually leads to errors when trying to execute it on the database.";
            throw new DbException(json_encode($errorMsg), ERROR_PLAY_SQL_FILE);
        }
        $this->db->playMultiQuery($query);
        try {
            $functionsHashTest = $this->db->sql('SELECT GETFUNCTIONSHASH() AS res;', ROW_EXPECT_1)['res'];
        } catch (DbException $e) {
            $functionsHashTest = null;
        }
        if ($functionsHash !== null AND $functionsHashTest === $functionsHash) {
            return $functionsHash;
        } else {
            $errorMsg[ERROR_MESSAGE_TO_USER] = 'Error while updating qfq.';
            $errorMsg[ERROR_MESSAGE_TO_DEVELOPER] = "Error while updating qfq: " .
                "Could not properly execute the file 'functions.sql' on the QFQ MYSQL database. " .
                "Please do so manually or set the parameter 'functionsHash' in the table comments of the table" .
                " 'Form' to 'no' i.e. 'Version=x.y.z&functionHash=no' to disable updating of the qfq mysql functions." .
                " You may also try giving the mysql user the 'SUPER' privilege temporarily.";
            throw new DbException(json_encode($errorMsg), ERROR_PLAY_SQL_FILE);
        }
    }

    /**
     * @param $path
     *
     * @return array
     * @throws CodeException
     */
    private function readUpdateData($path) {

        if (!@file_exists($path)) {
            throw new CodeException("File '$path'' not found", ERROR_IO_OPEN);
        }

        $UPDATE_ARRAY = null;

        include $path;

        return $UPDATE_ARRAY;

    }

    /**
     * Play all update statement with version number are '>' than $old and '<=' to $new.
     *
     * @param $old
     * @param $new
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    private function dbUpdateStatements($old, $new) {
        if ($new == '' || $old === false || $old === null) {
            return;
        }
        $updateArray = $this->readUpdateData(__DIR__ . '/DatabaseUpdateData.php');

        $apply = false;
        foreach ($updateArray as $key => $sqlStatements) {

            // Search starting point to apply updates. Do not apply updates for $key>$new
            $rc1 = version_compare($key, $old);
            $rc2 = version_compare($key, $new);
            if ($rc1 == 1 && $rc2 != 1) {
                $apply = true;
            }

            if ($apply) {
                // Play Statements
                foreach ($sqlStatements as $sql) {
                    $this->db->sql($sql, ROW_REGULAR, array(), "Apply updates to QFQ database. Installed version: $old. New QFQ version: $new");
                }
                // Remember already applied updates - in case something breaks and the update has to be repeated.
                $this->setDatabaseVersion($new);
            }
        }
    }
}